#include "sudoku.hpp"
#include <fstream>
#include <iostream>
#include <string>


Sudoku::Sudoku(void) {
    board.resize(9, std::vector<int>(9, -1));
}

void Sudoku::FillBoard(std::string filePath) {
    std::ifstream sudokuFile(filePath);
    if (!sudokuFile) {
        std::cerr << "Error opening file: " << filePath << std::endl;
        exit(1);
    }
    std::string line;
    int height = 0;
    while (!sudokuFile.eof() && height < 9) { 
        sudokuFile >> line;
        for (int i = 0; i < 9; ++i) { board[height][i] = line[i] - '0'; }
        ++height;
    }
}

// Base global call for solving
void Sudoku::Solve(void) {
    _Solve();
}

bool Sudoku::IsBoardSolved(void) {
    for (int i = 0; i < 9; ++i) {
        if (!IsColumnComplete(i)) { return false; }
    }
    return true;
}

void Sudoku::Print(void) {
    std::cout << "===================" << std::endl;
    for (auto i : board) {
        std::cout << '|';
        for (auto j : i) {
            std::cout << j << '|';
        }
        std::cout << std::endl;
    }
    std::cout << "===================" << std::endl;
}

// Recursive call for solving
bool Sudoku::_Solve(void) {
    int column, row;
    if (!FindEmptyCell(&column, &row)) { return true; }
    for (int i : GetUnusedRow(row)) {
        for (int j : GetUnusedColumn(column)) {
            for (int k : GetUnusedSubgrid(column, row)) {
                if (i == j == k) { 
                    board[row][column] = i; 
                    if (_Solve()) { return true; }
                    board[row][column] = 0;
                }
            }
        }
    }
    return 0;
}

// Fills column and row of an empty cell and returns true
// If no empty cell is found, return false
bool Sudoku::FindEmptyCell(int* column, int* row) {
    for (*row = 0; *row < 9; ++(*row)) {
        for (*column = 0; *column < 9; ++(*column)) {
            if (board[*row][*column] == 0) { return true; }
        }
    }
    return false;
}

bool Sudoku::IsValidPlacement(int column, int row, int box) {
    // Check column and row
    for (int i = 0; i < 9; ++i) {
        if (board[row][i] == box || board[i][column] == box) {
            return false;
        }
    }
    
    // Check subgrid
    int startRow = row - row % 3;
    int startColumn = column - column % 3;
    for (int i = 0; i < 3; ++i) {
        for (int j = 0; j < 3; ++j) {
            if (board[startRow + i][startColumn + j] == box) { return false; }
        }
    }
    return true;
}

bool Sudoku::IsRowComplete(int row) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int box;
    for (int i = 0; i < 9; i++) {
        box = board[row][i];
        if (leftOver[box] == 0 && box != 0) {
            std::cerr << "Invalid assignment detected. Aborting..." << std::endl;
            exit(1);
        }
        if (leftOver[box] == box) { leftOver[box] = 0; }
    }
    for (auto it = leftOver.begin(); it != leftOver.end();) {
        if (*it == 0) { it = leftOver.erase(it); }
        else { ++it; }
    }
    if (leftOver.size() != 0) { return false; }
    return true;
}

bool Sudoku::IsColumnComplete(int column) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int box;
    for (int i = 0; i < 9; i++) {
        box = board[i][column];
        if (leftOver[box] == 0 && box != 0) {
            leftOver[box] = -1; // Duplicate detected in same column
        }
        if (leftOver[box] == box) { leftOver[box] = 0; }
    }
    for (auto it = leftOver.begin(); it != leftOver.end();) {
        if (*it == 0) { it = leftOver.erase(it); }
        else { ++it; }
    }
    if (leftOver.size() != 0) { return false; }
    return true;
}

std::vector<int> Sudoku::GetUnusedRow(int row) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int box;
    for (int i = 0; i < 9; i++) {
        box = board[row][i];
        if (leftOver[box] == box) { leftOver[box] = 0; }
    }
    for (auto it = leftOver.begin(); it != leftOver.end();) {
        if (*it == 0) { it = leftOver.erase(it); }
        else { ++it; }
    }
    return leftOver;
}

std::vector<int> Sudoku::GetUnusedColumn(int column) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int box;
    for (int i = 0; i < 9; i++) {
        box = board[i][column];
        if (leftOver[box] == box) { leftOver[box] = 0; }
    }
    for (auto it = leftOver.begin(); it != leftOver.end();) {
        if (*it == 0) { it = leftOver.erase(it); }
        else { ++it; }
    }
    return leftOver;
}

std::vector<int> Sudoku::GetUnusedSubgrid(int column, int row) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int box;
    int startRow = row - row % 3;
    int startColumn = column - column % 3;
    for (int i = 0; i < 3; ++i) {
        for (int j = 0; j < 3; ++j) {
            box = board[row][column];
            if (board[startRow + i][startColumn + j] == box) { 
                leftOver[box] = 0;
            }
        }
    }
    for (auto it = leftOver.begin(); it != leftOver.end();) {
        if (*it == 0) { it = leftOver.erase(it); }
        else { ++it; }
    }
    return leftOver;
}
